from collections import OrderedDict
from datetime import datetime
import re

from django.utils import timezone
from zeep.helpers import serialize_object

from zeep.xsd import SkipValue, CompoundValue


def get_bool(value):
    return 'TRUE' if value else 'FALSE'


def get_optional_bool(value):
    return 'TRUE' if value else SkipValue


def convert_status(status):
    if isinstance(status, bool):
        return status

    return status.lower() == 'ok'


def camel_2_py(the_dict):
    new_dict = {}
    for key, val in the_dict.items():

        if len(re.sub(r"([A-Z])", r" \1", key).split()) == len(key):
            parts = [key.lower()]

        else:
            key = key.replace('ID', 'Id').replace('CRL', 'Crl')
            parts = re.sub(r"([A-Z])", r" \1", key).split()

        new_dict["_".join([x.lower() for x in parts])] = val

    return new_dict


def convert_time(timestamp):
    if isinstance(timestamp, datetime):
        if timezone.is_naive(timestamp):
            timestamp = timezone.make_aware(timestamp, timezone.utc)

        return timestamp

    return timezone.make_aware(datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%SZ"), timezone.utc)


def get_instance_converter(cls, prepare_kwargs=camel_2_py):
    def _convert_instance(value):
        if isinstance(value, CompoundValue):
            value = serialize_object(value)

        if isinstance(value, (dict, OrderedDict)):
            return cls(**prepare_kwargs(value))

        return value

    return _convert_instance


def get_typed_list_validator(klass):
    def _get_typed_list_validator(inst, attr, value):
        if not isinstance(value, list):
            raise TypeError('Value MUST be a list')

        if not all(isinstance(x, klass) for x in value):
            raise TypeError('Value MUST be a list of {}'.format(klass))

    return _get_typed_list_validator


def get_typed_list_converter(klass):
    converter = get_instance_converter(klass)

    def _get_typed_list_converter(value):
        return [converter(x) for x in value]

    return _get_typed_list_converter



from Crypto.PublicKey import ECC
from Crypto.PublicKey import RSA
from Crypto.Signature import DSS, PKCS1_v1_5


class PrehashedMessageData(object):
    """
    Since ECC.verify expects an hash but mobile id does verification without hashing we need
     to use a Hashlib compatible wrapper class for our data
    """
    def __init__(self, data):
        super(PrehashedMessageData, self).__init__()

        self._data = data

    def update(self, data):
        self._data = data

    def digest(self):
        return self._data


def verify_mid_signature(certificate_data, sp_challenge, response_challenge, signature):
    """ Verify mobile id Authentication signature is valid

    :param certificate_data: binary certificate data, from 'CertificateData' field
    :param sp_challenge: binary challenge sent via 'SPChallenge' field
    :param response_challenge: response challenge, from 'Challenge' field
    :param signature: response signature
    :return:
    """

    if not response_challenge.startswith(sp_challenge):
        return False

    try:
        key = RSA.importKey(certificate_data)
        verifier = PKCS1_v1_5.new(key)

    except ValueError:
        key = ECC.import_key(certificate_data)
        verifier = DSS.new(key, 'deterministic-rfc6979')

    digest = PrehashedMessageData(response_challenge)

    try:
        verifier.verify(digest, signature)

        return True

    except ValueError:
        return False



import binascii

from .digidocservice.service import DigiDocError, DigiDocNotOk


class BaseAction(object):
    @classmethod
    def do_action(cls, view, action_kwargs):
        raise NotImplementedError


class NoAction(object):
    @classmethod
    def do_action(cls, view, action_kwargs):
        return {'success': True}


class IdCardPrepareAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        service = view.get_service()

        if not view.get_files():
            return {
                'success': False,
                'code': 'MIN_1_FILE',
            }

        # Create signed document
        service.create_signed_document()

        # add all files
        for file in view.get_files():
            service.add_datafile(file.file_name, file.mimetype, service.HASHCODE, file.size, file.content)

        # Call prepare_signature
        try:
            response = service.prepare_signature(**action_kwargs)
            response['success'] = True
            return response

        except DigiDocNotOk as e:
            return {
                'success': False,
                'error_code': e.code,
                'message': str(e),
            }

        except DigiDocError as e:
            return {
                'success': False,
                'error_code': e.error_code,
                'message': service.ERROR_CODES.get(int(e.error_code), service.ERROR_CODES[100])
            }


class IdCardFinishAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        service = view.get_service()

        try:
            # Should we also pass back the signed_doc_info (?)
            service.finalize_signature(**action_kwargs)

            return {
                'success': True,
            }

        except DigiDocNotOk as e:
            return {
                'success': False,
                'error_code': e.code,
                'message': str(e),
            }

        except DigiDocError as e:
            return {
                'success': False,
                'error_code': e.error_code,
                'message': service.ERROR_CODES.get(int(e.error_code), service.ERROR_CODES[100])
            }


class SignCompleteAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        service = view.get_service()

        return service.get_file_data(view.get_files())


class MobileIdSignAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        service = view.get_service()

        if not view.get_files():
            return {
                'success': False,
                'code': 'MIN_1_FILE',
            }

        # Create signed document
        service.create_signed_document()

        # add all files
        for file in view.get_files():
            service.add_datafile(file.file_name, file.mimetype, service.HASHCODE, file.size, file.content)

        try:
            # Call sign
            resp = service.mobile_sign(**action_kwargs)

        except DigiDocError as e:
            return {
                'success': False,
                'error_code': e.error_code,
                'message': service.ERROR_CODES.get(int(e.error_code), service.ERROR_CODES[100])
            }

        return {
            'success': True,
            'challenge': resp['ChallengeID'],
        }


class MobileIdStatusAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        service = view.get_service()

        status_info = service.get_status_info()

        # If error occured
        if status_info['StatusCode'] not in ['OUTSTANDING_TRANSACTION', 'SIGNATURE', 'REQUEST_OK']:
            return {
                'success': False,
                'code': status_info['StatusCode'],
                'message': service.MID_STATUS_ERROR_CODES[status_info['StatusCode']],
            }

        elif status_info['StatusCode'] == 'OUTSTANDING_TRANSACTION':
            return {
                'success': False,
                'pending': True,
            }

        else:
            return {
                'success': True,
            }


class MobileIdAuthenticateAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        service = view.get_service()

        try:
            # Call mobile_authenticate
            resp, challenge = service.mobile_authenticate(**action_kwargs)

            resp_challenge = binascii.unhexlify(resp['Challenge'])

            # Modify stored digidoc session
            view.set_digidoc_session(service.session_code)

            full_name = ' '.join([resp['UserGivenname'], resp['UserSurname']]).title()

            # Store Certificate owner information in session
            view.set_digidoc_session_data('mid_id_code', resp['UserIDCode'])
            view.set_digidoc_session_data('mid_firstname', resp['UserGivenname'])
            view.set_digidoc_session_data('mid_lastname', resp['UserSurname'])
            view.set_digidoc_session_data('mid_full_name', full_name)
            view.set_digidoc_session_data('mid_country', resp['UserCountry'])
            view.set_digidoc_session_data('mid_common_name', resp['UserCN'])

            # Store CertificateData in session (so we can verify later based on it)
            view.set_digidoc_session_data('mid_sp_challenge', challenge)
            view.set_digidoc_session_data('mid_resp_challenge', resp_challenge)
            view.set_digidoc_session_data('mid_certificate_data', resp['CertificateData'])

        except DigiDocError as e:
            return {
                'success': False,
                'error_code': e.error_code,
                'message': service.ERROR_CODES.get(int(e.error_code), service.ERROR_CODES[100])
            }

        return {
            'success': True,
            'challenge': resp['ChallengeID'],
            'challenge_raw': resp['Challenge'],
        }


class MobileIdAuthenticateStatusAction(BaseAction):
    @classmethod
    def do_action(cls, view, action_kwargs):
        service = view.get_service()

        try:
            status_code, signature = service.get_mobile_authenticate_status(**action_kwargs)

        except DigiDocError as e:
            return {
                'success': False,
                'pending': False,
                'code': e.error_code,
                'message': e.known_fault,
            }

        # FIXME: After signature verification is added, make sure to verify the signature here

        # If an error occurred
        if status_code not in ['OUTSTANDING_TRANSACTION', 'USER_AUTHENTICATED']:
            return {
                'success': False,
                'pending': False,
                'code': status_code,
                'message': service.MID_STATUS_ERROR_CODES[status_code],
            }

        if status_code == 'OUTSTANDING_TRANSACTION':
            return {
                'success': False,
                'pending': True,
                'code': status_code,
                'message': None,
            }

        # signature

        sp_challenge = view.get_digidoc_session_data('mid_sp_challenge')
        resp_challenge = view.get_digidoc_session_data('mid_resp_challenge')
        certificate_data = view.get_digidoc_session_data('mid_certificate_data')

        if not service.verify_mid_signature(certificate_data, sp_challenge, resp_challenge, signature):
            return {
                'success': False,
                'pending': False,
                'code': 'BAD_SIGNATURE',
                'message': service.MID_STATUS_ERROR_CODES['BAD_SIGNATURE'],
            }

        # USER_AUTHENTICATED
        return {
            'success': True,
            'pending': False,
            'code': status_code,
            'message': None,
        }



import os
import subprocess
import logging

from tempfile import NamedTemporaryFile

from django.utils.encoding import force_bytes, force_text


logger = logging.getLogger(__name__)


class OCSPVerifier(object):
    CERT_PATH = os.path.join(os.path.dirname(__file__), 'certs')

    ISSUER_CERTS = {
        # Test certs
        'TEST of SK OCSP RESPONDER 2011': 'TEST_OCSP_2011.pem',
        'TEST of EID-SK 2011': 'TEST_of_EID-SK_2011.pem',
        'TEST of EID-SK 2016': 'TEST_of_EID-SK_2016.pem',
        'TEST of ESTEID-SK 2011': 'TEST_of_ESTEID-SK_2011.pem',
        'TEST of ESTEID-SK 2015': 'TEST_of_ESTEID-SK_2015.pem',
        'TEST of KLASS3-SK 2010': 'TEST_of_KLASS3-SK_2010.pem',
        'TEST of KLASS3-SK 2016': 'TEST_of_KLASS3-SK_2016.pem',

        # Live certs
        'ESTEID-SK 2007': 'ESTEID-SK_2007.pem',
        'ESTEID-SK 2011': 'ESTEID-SK_2011.pem',
        'ESTEID-SK 2015': 'ESTEID-SK_2015.pem',
        'EID-SK 2011': 'EID-SK_2011.pem',
        'EID-SK 2016': 'EID-SK_2016.pem',
        'KLASS3-SK 2010': 'KLASS3-SK_2010_EECCRCA.pem',
        'KLASS3-SK 2016': 'KLASS3-SK_2016_EECCRCA_SHA384.pem',
    }

    def __init__(self, certificate, issuer, ocsp_url, va_file_path):
        self.certificate = certificate
        self.issuer = issuer
        if isinstance(self.issuer, str):
            self.issuer = self.parse_issuer(issuer)

        self.ocsp_url = ocsp_url
        self.va_file_path = va_file_path

    def verify(self):
        if not self.certificate:
            return 1

        if not self.issuer:
            return 2

        issuer_cn = self.issuer.get('CN', None)
        if not issuer_cn:
            return 3

        issuer_cert_filename = self.get_issuer_cert_filename(issuer_cn)
        if issuer_cert_filename is None:
            return 4

        # Save cert to a temporary file
        cert_file = NamedTemporaryFile(delete=False)
        cert_file.write(force_bytes(self.certificate.replace('\t', '')))
        cert_file.close()

        # Run openssl ocsp verify command
        args = [
            'openssl', 'ocsp',
            '-issuer %s' % issuer_cert_filename,
            '-cert %s' % cert_file.name,
            '-url %s' % self.ocsp_url,
            '-VAfile %s' % self.va_file_path,
        ]

        try:
            output = subprocess.check_output(' '.join(args), stderr=subprocess.STDOUT, shell=True)
            output = force_text(output)

            if ': good' in output:
                result = 0

            elif ': revoked' in output:
                result = 6

            else:
                logger.info('openssl ocsp: unknown output: %s', output)
                result = 7

        except subprocess.CalledProcessError as e:
            logger.error('openssl ocsp: failed with output: %s', e.output)
            result = 5

        # delete temporary file
        os.unlink(cert_file.name)

        return result

    @classmethod
    def parse_issuer(cls, issuer):
        res = {}
        issuer = issuer.strip().strip('/').split('/')

        for part in issuer:
            part = part.split('=')
            res[part[0]] = part[1]

        return res

    def get_issuer_cert_filename(self, issuer_cn):
        filename = self.ISSUER_CERTS.get(issuer_cn, None)
        if filename is None:
            return None

        return os.path.join(self.CERT_PATH, filename)



import re
import sys

from platform import python_version_tuple

from six import unichr


def ucs_to_utf8(val):
    klass = bytes

    # For py2:
    #  - bytes is an alias for str: use bytearray instead
    if sys.version_info[0] < 3:
        klass = bytearray

    return klass([ord(x) for x in re.sub(r"\\([0-9ABCDEF]{1,2})",
                                         lambda x: unichr(int(x.group(1), 16)), val)]).decode('utf-8')


def get_name_from_legacy_common_name(common_name):
    common_name = common_name.replace('\\,', ',')
    common_name = common_name.strip().rstrip(',')

    parts = common_name.split(',')[:-1]
    parts.reverse()

    return ' '.join(parts).title()


def get_id_from_legacy_common_name(common_name):
    common_name = common_name.strip().rstrip(',')

    return common_name.split(',')[-1]


def parse_legacy_dn(dn):
    x_client = dn.strip().strip('/').split('/')

    res = {}

    for part in x_client:
        part = ucs_to_utf8(part).split('=')
        res[part[0]] = part[1]

    return res


def parse_rfc_dn(dn):
    dn = ucs_to_utf8(dn).replace('\\,', ',')
    res = {}
    c_key = None

    for part in dn.strip().split(','):
        if '=' in part:
            part = part.split('=')

            c_key = part[0]
            res[c_key] = part[1]

        elif c_key:
            res[c_key] = u'{0},{1}'.format(res[c_key], part)

    return res


major, minor, _ = python_version_tuple()

#  py 2.x: where pubkey_bytes is a wrapped str
if major == '2':  # pragma: no cover
    def get_hex_from_bytes(buf):
        return buf.encode('hex')

#  py 3.4: since bytes.hex was added in 3.5
elif major == '3' and int(minor) <= 4:  # pragma: no cover
    import codecs

    def get_hex_from_bytes(buf):
        return codecs.encode(buf, 'hex_codec')

else:
    def get_hex_from_bytes(buf):
        return buf.hex()



from . import config
from .actions import (IdCardPrepareAction, IdCardFinishAction, SignCompleteAction, MobileIdSignAction,
                      MobileIdStatusAction, BaseAction, MobileIdAuthenticateAction,
                      MobileIdAuthenticateStatusAction)
from .digidocservice.service import DigiDocService, DigiDocError
from .response import JSONResponse


class GenericDigitalSignViewMixin(object):
    autostart_digidoc_session = True

    DIGIDOC_SESSION_KEY = '__ddoc_session'
    DIGIDOC_SESSION_DATA_KEY = '__ddoc_session_data'

    def get_files(self):
        """ This should be implemented on view level, and should return a list
            of files that should be digitally signed
        """
        return []

    def set_digidoc_session(self, session_code):
        """ Stores the new session token so it is remembered across requests.

        """
        self.request.session[self.DIGIDOC_SESSION_KEY] = session_code

    def set_digidoc_session_data(self, key, value):
        session_data = self.request.session.get(self.DIGIDOC_SESSION_DATA_KEY, {})
        session_data.update({key: value})

        self.request.session[self.DIGIDOC_SESSION_DATA_KEY] = session_data

    def get_digidoc_session_data(self, key):
        session_data = self.request.session.get(self.DIGIDOC_SESSION_DATA_KEY, {})

        return session_data.get(key, None)

    def get_digidoc_session(self):
        """ This method returns the active digidoc session
            associated with the active request.

            DigiDocService session is stored in request.session[I{DIGIDOC_SESSION_KEY}]
        """

        return self.request.session.get(self.DIGIDOC_SESSION_KEY)

    def destroy_digidoc_session(self):
        """ Closes DigiDocService session and clears request.session[I{DIGIDOC_SESSION_KEY}]
        """

        # cleanup data too
        self.destroy_digidoc_session_data()

        try:
            session = self.request.session[self.DIGIDOC_SESSION_KEY]

            if session:
                try:
                    service = self.flat_service()
                    service.session_code = session
                    service.close_session()

                except DigiDocError:
                    pass

            del self.request.session[self.DIGIDOC_SESSION_KEY]

        except KeyError:
            pass

    def destroy_digidoc_session_data(self):
        # Ensure hanging references to the data are cleared and then remove it from session
        self.request.session.get(self.DIGIDOC_SESSION_DATA_KEY, {}).clear()
        self.request.session.pop(self.DIGIDOC_SESSION_DATA_KEY, {})

    def flat_service(self):
        return DigiDocService(
            wsdl_url=config.wsdl_url(),
            service_name=config.service_name(),
            mobile_message=config.mobile_message(),
        )

    def start_session_kwargs(self):
        return {
            'b_hold_session': True,
        }

    def get_service(self):
        if hasattr(self, 'stored_service'):
            return self.stored_service

        service = self.flat_service()
        session = self.get_digidoc_session()

        if session:
            service.session_code = session

        else:
            if self.autostart_digidoc_session:
                if service.start_session(**self.start_session_kwargs()):
                    # Clear session data to avoid data leakage
                    self.destroy_digidoc_session_data()

                    # Store the new session identifier
                    self.set_digidoc_session(service.session_code)

        setattr(self, 'stored_service', service)
        return service


class BaseMobileIdAuthenticateViewMixin(GenericDigitalSignViewMixin):
    # MobileAuthenticate starts session automatically and does not accept Sesscode
    autostart_digidoc_session = False


class BaseDigitalSignViewMixin(GenericDigitalSignViewMixin):
    CATCH_POST = True
    ACTION_CLASS = None

    def post(self, request, *args, **kwargs):
        if self.CATCH_POST:
            return JSONResponse(self.do_action(*args, **kwargs))

        return super(BaseDigitalSignViewMixin, self).post(request, *args, **kwargs)

    def build_action_kwargs(self):
        return {}

    def do_action(self, *args, **kwargs):
        if not self.ACTION_CLASS or not issubclass(self.ACTION_CLASS, BaseAction):
            raise NotImplementedError

        return self.ACTION_CLASS.do_action(self, action_kwargs=self.build_action_kwargs())


class IdCardPrepareViewMixin(BaseDigitalSignViewMixin):
    ACTION_CLASS = IdCardPrepareAction


class IdCardFinishViewMixin(BaseDigitalSignViewMixin):
    ACTION_CLASS = IdCardFinishAction


class DigidocCompleteViewMixin(BaseDigitalSignViewMixin):
    """ This action gets the signed document from this DigiDocService session
        and then closes it. It's up to the user what he/she wants to do with it.
    """
    CATCH_POST = False
    ACTION_CLASS = SignCompleteAction


class MobileIdSignViewMixin(BaseDigitalSignViewMixin):
    ACTION_CLASS = MobileIdSignAction


class MobileIdStatusViewMixin(BaseDigitalSignViewMixin):
    ACTION_CLASS = MobileIdStatusAction


class MobileIdAuthenticateViewMixin(BaseDigitalSignViewMixin, BaseMobileIdAuthenticateViewMixin):
    ACTION_CLASS = MobileIdAuthenticateAction


class MobileIdAuthenticateStatusViewMixin(BaseDigitalSignViewMixin):
    ACTION_CLASS = MobileIdAuthenticateStatusAction



import os

from django.conf import settings


def wsdl_url():
    """Url of the DigidocService wsdl

    Test: https://tsp.demo.sk.ee/dds.wsdl
    Live: https://digidocservice.sk.ee/dds.wsdl

    :return: str
    """
    return getattr(settings, 'DIGIDOC_SERVICE_WSDL_URL', 'https://tsp.demo.sk.ee/dds.wsdl')


def service_name():
    return getattr(settings, 'DIGIDOC_SERVICE_NAME', 'Testimine')


def mobile_message():
    return getattr(settings, 'DIGIDOC_SERVICE_MESSAGE', 'Testimine')


def get_hosts():
    return getattr(settings, 'HOSTS', {})


def ocsp_url():
    """OCSP server url

    Test: http://demo.sk.ee/ocsp
    Live: http://ocsp.sk.ee

    :return: str
    """
    return getattr(settings, 'ESTEID_OCSP_URL', 'http://demo.sk.ee/ocsp')


def ocsp_responder_certificate_path():
    """Get ocsp responder certificate path

    Test: TEST_of_SK_OCSP_RESPONDER_2011.pem
    Live: sk-ocsp-responder-certificates.pem

    Note: These files are distributed under esteid/certs

    :return:
    """
    certificate_path = getattr(settings, 'ESTEID_OCSP_RESPONDER_CERTIFICATE_PATH', 'TEST_of_SK_OCSP_RESPONDER_2011.pem')

    if certificate_path in ['TEST_of_SK_OCSP_RESPONDER_2011.pem', 'sk-ocsp-responder-certificates.pem']:
        return os.path.join(os.path.dirname(__file__), 'certs', certificate_path)

    return certificate_path



from datetime import datetime

import attr

from esteid.digidocservice.util import (convert_time, convert_status, get_typed_list_validator, get_typed_list_converter,
                                        get_instance_converter, camel_2_py)
from esteid.helpers import get_name_from_legacy_common_name


class FromDictMixin(object):
    @classmethod
    def from_dict(cls, the_dict):
        return cls(**camel_2_py(the_dict))


@attr.s
class CertificatePolicy(FromDictMixin):
    oid = attr.ib()
    url = attr.ib(default=None)
    description = attr.ib(default=None)


@attr.s
class Certificate(FromDictMixin):
    valid_from = attr.ib()  # this might be a timestamp
    issuer_serial = attr.ib()
    issuer = attr.ib()
    valid_to = attr.ib()  # this might be a timestamp
    subject = attr.ib()

    policies = attr.ib(default=attr.Factory(list), validator=[
        attr.validators.instance_of(list),
        get_typed_list_validator(CertificatePolicy),
    ], converter=get_typed_list_converter(CertificatePolicy))


@attr.s
class ResponderCertificate(Certificate):
    pass


@attr.s
class Signer(FromDictMixin):
    id_code = attr.ib()

    # Note: This should also support common_name argument coming in
    full_name = attr.ib()

    certificate = attr.ib(validator=attr.validators.instance_of(Certificate),
                          converter=get_instance_converter(Certificate))

    @staticmethod
    def prepare_kwargs(kwargs):
        kwargs = camel_2_py(kwargs)
        full_name = kwargs.get('full_name', None)

        # If the full_name is None, check common_name field and use that
        if full_name is None:
            full_name = get_name_from_legacy_common_name(kwargs.pop('common_name', None))

        kwargs['full_name'] = full_name

        return kwargs


@attr.s
class Confirmation(FromDictMixin):
    produced_at = attr.ib()  # this might be a timestamp
    responder_id = attr.ib()

    responder_certificate = attr.ib(validator=attr.validators.instance_of(ResponderCertificate),
                                    converter=get_instance_converter(ResponderCertificate))


@attr.s
class SignatureProductionPlace(FromDictMixin):
    postal_code = attr.ib()
    country_name = attr.ib()
    city = attr.ib()
    state_or_province = attr.ib()


@attr.s
class SignerRole(FromDictMixin):
    role = attr.ib()
    certified = attr.ib()


@attr.s
class SignatureInfo(FromDictMixin):
    signing_time = attr.ib(validator=attr.validators.instance_of(datetime), converter=convert_time)
    status = attr.ib(converter=convert_status)
    id = attr.ib()

    signer = attr.ib(validator=attr.validators.instance_of(Signer),
                     converter=get_instance_converter(Signer, prepare_kwargs=Signer.prepare_kwargs))
    confirmation = attr.ib(validator=attr.validators.instance_of(Confirmation), converter=get_instance_converter(Confirmation))

    signature_production_place = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(SignatureProductionPlace)),
                                         converter=attr.converters.optional(get_instance_converter(SignatureProductionPlace)), default=None)

    signer_role = attr.ib(default=attr.Factory(list), validator=[
        attr.validators.instance_of(list),
        get_typed_list_validator(SignerRole),
    ], converter=get_typed_list_converter(SignerRole))

    error = attr.ib(default=None)
    crl_info = attr.ib(default=None)

    # references Timestamps object in the docs, currently we do not parse it to a internal structure
    # http://sk-eid.github.io/dds-documentation/api/api_docs/#signeddocinfo
    timestamps = attr.ib(default=None)


@attr.s
class DataFileInfo(FromDictMixin):
    size = attr.ib()
    id = attr.ib()
    mime_type = attr.ib()
    content_type = attr.ib()
    digest_value = attr.ib()
    digest_type = attr.ib()
    filename = attr.ib()

    # Note: This is not documented as of 19:35 20.11.2017
    attributes = attr.ib(default=None)


@attr.s
class SignedDocInfo(FromDictMixin):
    format = attr.ib()
    version = attr.ib()

    data_file_info = attr.ib(default=attr.Factory(list), validator=[
        attr.validators.instance_of(list),
        get_typed_list_validator(DataFileInfo),
    ], converter=get_typed_list_converter(DataFileInfo))

    signature_info = attr.ib(default=attr.Factory(list), validator=[
        attr.validators.instance_of(list),
        get_typed_list_validator(SignatureInfo),
    ], converter=get_typed_list_converter(SignatureInfo))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys

import esteid

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

version = esteid.__version__

readme = open('README.md').read()

setup(
    name='django-esteid',
    version=version,
    description="""Django-esteid is a package that provides Esteid based authentication for your Django applications.""",
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Thorgate',
    author_email='jyrno@thorgate.eu',
    url='https://github.com/thorgate/django-esteid',
    packages=[
        'esteid',
    ],
    include_package_data=True,
    install_requires=[
        'django>=1.8,!=2.1.0,!=2.1.1',
        'attrs>=19.2.0',
        'lxml>=3.4,<=4.3.5',
        'zeep>=2.4.0',
        'pycryptodome>=3.7.2',
    ],
    license="BSD",
    zip_safe=False,
    keywords='esteid django',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



# -*- coding: utf-8 -*-
import base64
import binascii
import logging
import os

from django.utils.encoding import force_text, force_bytes

from zeep import Client, Transport
from zeep.cache import SqliteCache
from zeep.exceptions import Fault
from zeep.helpers import serialize_object
from zeep.xsd import SkipValue

try:
    from zeep.settings import Settings as ZeepSettings

except ImportError:  # pragma: no cover
    # only occurs on zeep before 3.0
    ZeepSettings = None

from ..signature import verify_mid_signature

from .containers import BdocContainer
from .types import SignedDocInfo
from .util import get_bool, get_optional_bool


class DigiDocException(Exception):
    """ Unknown errors
    """

    def __init__(self, command, params, *args):
        self.command = command
        self.params = params

        super(DigiDocException, self).__init__(*args)


class DigiDocError(Exception):
    """ Known errors
    """

    def __init__(self, error_code, known_fault, *args):
        self.error_code = error_code
        self.known_fault = known_fault

        super(DigiDocError, self).__init__(*args)


class DigiDocNotOk(Exception):
    """ Used when an api is supposed to return status OK but does not
    """

    code = -1

    def __init__(self, response, message):
        self.response = response

        super(Exception, self).__init__(message)


class PreviouslyCreatedContainer(object):
    pass


class DataFile(object):
    def __init__(self, file_name, mimetype, content_type, size, content, info=None):
        self.file_name = file_name
        self.mimetype = mimetype
        self.content_type = content_type or DigiDocService.HASHCODE
        self.size = size
        self.content = content

        self.info = info


class DigiDocService(object):
    RESPONSE_STATUS_OK = "OK"

    # FIXME: error i18n (allow to set language in init, can be overwritten in mobile_authenticate/mobile_sign)
    ERROR_CODES = {
        -1: 'Päring ebaõnnestus, staatus: %s',
        100: 'Üldine viga.',
        101: 'Vigased sissetulevad parameetrid.',
        102: 'Mõned sissetulevad parameetrid on puudu.',
        103: 'Teenuse omanikul puudub õigus teha päringuid allkirja-kontrolli teenusesse (OCSP: AUTORISEERIMATA)',
        104: 'Teenusele juurdepääs pole lubatud.',
        105: 'Kasutajal puuduvad küsitud tüüpi sertifikaadid.',
        200: 'Üldine teenuse viga.',
        201: 'Kasutaja sertifikaat on puudu.',
        202: 'Sertifikaadi korrektsust polnud võimalik valideerida.',
        203: 'Sessioon on lukustatud teise SOAPi pärginu poolt.',
        300: 'Üldine viga seoses kasutaja telefoniga.',
        301: 'Pole Mobiil-ID kasutaja.',
        302: 'Sertifikaat ei kehti (OCSP: TAGASI VÕETUD).',
        303: 'Sertifikaat ei ole aktiveeritud ja/ või selle staatus on teadmata (OCSP: TEADMATA).',
        304: 'Sertifikaat on peatatud.',
        305: 'Sertifikaat on aegunud.',
        413: 'Sissetulev päring ületab teenuse lubatud mahupiiranguid.',
        503: 'Teenuse üheaegselt esitatud päringute piirang on ületatud.',
    }

    MID_STATUS_ERROR_CODES = {
        'EXPIRED_TRANSACTION': 'MobiilID allkirjastamise ajapiirang sai läbi.',
        'USER_CANCEL': 'Kasutaja katkestas allkirjastamise.',
        'NOT_VALID': 'Allkiri ei kehti.',
        'MID_NOT_READY': 'Mobiil-ID ei ole veel sellel telefonil aktiveeritud. Palun proovige hiljem uuesti.',
        'PHONE_ABSENT': 'Telefon ei ole kättesaadav.',
        'SENDING_ERROR': 'Ei suutnud telefonile Mobiil-ID päringut saata.',
        'SIM_ERROR': 'Telefoni SIM-kaardiga tekkis probleem.',
        'OCSP_UNAUTHORIZED': 'Mobiil-ID kasutajal ei ole lubatud teha OSCP päringuid.',
        'INTERNAL_ERROR': 'Serveri viga Mobiil-ID allkirjastamisel.',
        'REVOKED_CERTIFICATE': 'Allkirjastaja sertifikaat ei kehti.',
        'BAD_SIGNATURE': 'Signatuuri valideerimine ebaõnnestus.',
    }

    LANGUAGE_ET = 'EST'
    LANGUAGE_EN = 'ENG'
    LANGUAGE_RU = 'RUS'
    LANGUAGE_LT = 'LIT'

    HASHCODE = 'HASHCODE'
    EMBEDDED_BASE64 = 'EMBEDDED_BASE64'

    # Commands which don't need Sesscode as input
    SESSION_INIT_COMMANDS = [
        'StartSession',
        'MobileAuthenticate',
    ]

    def __init__(self, wsdl_url, service_name, mobile_message='Signing via python', transport=None):
        self.service_name = service_name
        self.mobile_message = mobile_message

        self.session_code = None
        self.data_files = []
        self.container = None

        if wsdl_url == 'https://tsp.demo.sk.ee/dds.wsdl':  # pragma: no branch
            assert service_name == 'Testimine', 'When using Test DigidocService the service name must be `Testimine`'

        the_transport = transport or Transport(cache=SqliteCache())

        if ZeepSettings is not None:  # pragma: no branch
            settings = ZeepSettings(strict=False)

            self.client = Client(wsdl_url, transport=the_transport, settings=settings)

        else:
            self.client = Client(wsdl_url, transport=the_transport, strict=False)

    def start_session(self, b_hold_session, sig_doc_xml=None, datafile=None):
        """Start a DigidocService session

        :return: True if session was started and session code was stored in I{session_code}
        """
        response = self.__invoke('StartSession', {
            'bHoldSession': b_hold_session,
            'SigDocXML': sig_doc_xml or SkipValue,
            'datafile': datafile or SkipValue,

            # This parameter is deprecated and exists only due to historical reasons. We need to specify it as
            #  SkipValue to keep zeep happy
            'SigningProfile': SkipValue,
        })

        if response['Sesscode']:
            self.data_files = []
            self.session_code = response['Sesscode']

            if sig_doc_xml:
                self.container = PreviouslyCreatedContainer()

            return True

        # If b_hold_session is set to False, response will not contain a session
        # in case of errors, exceptions are raised from __invoke anyway
        return False

    def mobile_authenticate(self, id_code, country, phone_nr, message=None, language=None,
                            return_cert_data=True, return_revocation_data=False):
        challenge = self.get_sp_challenge()

        response = self.__invoke('MobileAuthenticate', {
            # allow to not define IDCode - it's only required w/ Lithuanian Mobile-ID
            # see section 7.1 in http://www.sk.ee/upload/files/DigiDocService_spec_eng.pdf
            'IDCode': id_code or SkipValue,

            'CountryCode': country,

            'PhoneNo': phone_nr,
            'Language': self.parse_language(language),
            'ServiceName': self.service_name,
            'MessageToDisplay': message or self.mobile_message,
            'SPChallenge': force_text(binascii.hexlify(challenge)),

            'MessagingMode': 'asynchClientServer',
            'AsyncConfiguration': SkipValue,

            'ReturnCertData': get_optional_bool(return_cert_data),
            'ReturnRevocationData': get_optional_bool(return_revocation_data),
        }, no_raise=True)

        # Update session code
        if response['Sesscode']:  # pragma: no branch (only missing in older service versions)
            self.data_files = []
            self.session_code = response['Sesscode']

        if 'CertificateData' in response and response['CertificateData']:
            # certificate data is b64 encoded binary (not PEM)
            response['CertificateData'] = base64.b64decode(response['CertificateData'])

        else:
            response['CertificateData'] = None

        if 'RevocationData' in response and response['RevocationData']:
            # certificate data is b64 encoded binary (not PEM)
            response['RevocationData'] = base64.b64decode(response['RevocationData'])

        else:
            response['RevocationData'] = None

        return response, challenge

    def get_mobile_authenticate_status(self, wait=False):
        response = self.__invoke('GetMobileAuthenticateStatus', {
            'WaitSignature': get_bool(wait),
        }, no_raise=True)

        status_code, signature = response['Status'], None

        if 'Signature' in response and response['Signature']:
            signature = base64.b64decode(response['Signature'])

        return status_code, signature

    def create_signed_document(self, file_format='BDOC', signing_profile='LT_TM'):
        if self.container and isinstance(self.container, PreviouslyCreatedContainer):
            raise DigiDocException('CreateSignedDoc', {}, 'PreviouslyCreatedContainer already in session')

        versions = {
            'BDOC': '2.1',
        }
        containers = {
            'BDOC': BdocContainer,
        }

        if file_format not in versions:
            raise ValueError('File format should be one of: %s' % versions.keys())

        self.__invoke('CreateSignedDoc', {
            'Format': file_format,
            'Version': versions[file_format],

            # Either LT or LT_TM, see: http://sk-eid.github.io/dds-documentation/api/api_docs/#mobilesign
            'SigningProfile': signing_profile,
        })

        self.container = containers[file_format]

        return True

    def add_datafile(self, file_name, mimetype, content_type, size, content):
        if self.container and isinstance(self.container, PreviouslyCreatedContainer):
            raise DigiDocException('AddDataFile', {}, 'Cannot add files to PreviouslyCreatedContainer')

        assert self.container, 'Must create a signed document before adding files'
        assert content_type in [self.HASHCODE, self.EMBEDDED_BASE64], 'Invalid content_type'

        args = {
            'FileName': file_name,
            'MimeType': mimetype,
            'ContentType': content_type,
            'Content': SkipValue,
            'Size': size,

            'DigestType': SkipValue,
            'DigestValue': SkipValue,
        }

        # We only support HASHCODE format, see above
        if content_type == self.EMBEDDED_BASE64:
            args['Content'] = base64.b64encode(content)

        elif content_type == self.HASHCODE:
            args['DigestType'] = self.container.DEFAULT_HASH_ALGORITHM
            args['DigestValue'] = force_text(self.container.hash_code(content))

        response = self.__invoke('AddDataFile', args)

        # Get the file from the response info object
        info = next(iter([f for f in response['SignedDocInfo']['DataFileInfo'] if f['Filename'] == file_name]))

        self.data_files.append(DataFile(file_name, mimetype, content_type, size, content, info))

        return self.data_files

    def mobile_sign(self, id_code, country, phone_nr, language=None, signing_profile='LT_TM'):
        """ This can be used to add a signature to existing data files

            WARNING: Must have at least one datafile in the session
        """

        if not (self.container and isinstance(self.container, PreviouslyCreatedContainer)):
            assert self.data_files, 'To use MobileSign endpoint the application must ' \
                                    'add at least one data file to users session'

        response = self.__invoke('MobileSign', {
            'SignerIDCode': id_code,
            'SignersCountry': country,
            'SignerPhoneNo': phone_nr,
            'Language': self.parse_language(language),

            'Role': SkipValue,
            'City': SkipValue,
            'StateOrProvince': SkipValue,
            'PostalCode': SkipValue,
            'CountryName': SkipValue,

            'ServiceName': self.service_name,
            'AdditionalDataToBeDisplayed': self.mobile_message,

            # Either LT or LT_TM, see: http://sk-eid.github.io/dds-documentation/api/api_docs/#mobilesign
            'SigningProfile': signing_profile,

            'MessagingMode': 'asynchClientServer',
            'AsyncConfiguration': SkipValue,

            'ReturnDocInfo': SkipValue,
            'ReturnDocData': SkipValue,
        })

        return response

    def prepare_signature(self, certificate, token_id, role='', city='', state='', postal_code='', country='', signing_profile='LT_TM'):
        if not (self.container and isinstance(self.container, PreviouslyCreatedContainer)):
            assert self.data_files, 'To use PrepareSignature endpoint the application must ' \
                                    'add at least one data file to users session'

        response = self.__invoke('PrepareSignature', {
            'SignersCertificate': certificate,
            'SignersTokenId': token_id,
            'Role': role,
            'City': city,
            'State': state,
            'PostalCode': postal_code,
            'Country': country,
            # Either LT or LT_TM, see: http://sk-eid.github.io/dds-documentation/api/api_docs/#preparesignature
            'SigningProfile': signing_profile,
        })

        if response['Status'] == self.RESPONSE_STATUS_OK:
            return {
                'id': response['SignatureId'],
                'digest': response['SignedInfoDigest'],
            }

        raise DigiDocNotOk(response, self.ERROR_CODES[-1] % response['Status'])

    def finalize_signature(self, signature_id, signature_value):
        response = self._finalize_signature(signature_id, signature_value)

        if response['Status'] == self.RESPONSE_STATUS_OK:
            response_dict = serialize_object(response['SignedDocInfo'])

            return SignedDocInfo.from_dict(response_dict)

        raise DigiDocNotOk(response, self.ERROR_CODES[-1] % response['Status'])

    def close_session(self):
        response = self.__invoke('CloseSession')

        self.data_files = []
        self.session_code = None

        return response

    def get_signed_doc(self):
        response = self.__invoke('GetSignedDoc')

        if response['Status'] == self.RESPONSE_STATUS_OK:
            return base64.b64decode(force_bytes(response['SignedDocData']))

        raise DigiDocNotOk(response, self.ERROR_CODES[-1] % response['Status'])

    def get_signed_doc_info(self):
        response = self.__invoke('GetSignedDocInfo')
        response_dict = serialize_object(response['SignedDocInfo'])

        return SignedDocInfo.from_dict(response_dict)

    def get_status_info(self, wait=False):
        response = self.__invoke('GetStatusInfo', {
            'ReturnDocInfo': False,
            'WaitSignature': wait,
        })

        return response

    def __invoke(self, command, params=None, no_raise=False):
        params = params or {}

        if command not in self.SESSION_INIT_COMMANDS:
            params.update({'Sesscode': self.session_code})

        try:
            response = getattr(self.client.service, command)(**params)

            logging.debug('%s:Response: %s', command, response)

            if response == self.RESPONSE_STATUS_OK:
                return True

            elif response['Status'] == self.RESPONSE_STATUS_OK:
                return response

            elif no_raise:
                # Some service methods use the status field for other things (e.g. MobileAuthenticate)
                return response

            # This should usually not happen, hence the over-the-top raise Exception which gets re-raised as
            #  DigiDocException below
            raise Exception(response)

        except Fault as e:
            error_code = int(str(e))
            known_fault = self.ERROR_CODES.get(error_code, None)

            logging.debug('Response body [/%s - %s]: %s', command, error_code, e.message)

            if known_fault is not None:
                raise DigiDocError(error_code, known_fault,
                                   "Server result [/%s - %s]: %s" % (command, error_code, known_fault))

            else:
                logging.exception('Request to %s with params %s caused an error', command, params)
                raise DigiDocException(command, params, e)

        except Exception as e:
            logging.exception('Request to %s with params %s caused an error', command, params)

            raise DigiDocException(command, params, e)

    def get_file_data(self, the_files):
        # Add all files to memory
        for file in the_files:
            assert isinstance(file, DataFile)
            self.data_files.append(file)

        # Get bdoc container from DigidocService
        file_data = self.get_signed_doc()

        with self.to_bdoc(file_data) as container:
            file_data = container.data_files_format()

        return file_data

    def to_bdoc(self, file_data):
        return BdocContainer(file_data, self.data_files)

    def verify_mid_signature(self, certificate_data, sp_challenge, response_challenge, signature):
        return verify_mid_signature(certificate_data, sp_challenge, response_challenge, signature)

    def get_sp_challenge(self):
        return os.urandom(10)

    def parse_language(self, language):
        if language is None:
            return self.LANGUAGE_ET

        assert language in [self.LANGUAGE_ET, self.LANGUAGE_EN, self.LANGUAGE_RU, self.LANGUAGE_LT]

        return language

    def _finalize_signature(self, signature_id, signature_value):
        return self.__invoke('FinalizeSignature', {
            'SignatureId': signature_id,
            'SignatureValue': signature_value,
        })
